#!/usr/bin/perl

#
# iet interface to /proc/net/iet
#
# by Steffen Plotner, 2007, 2010
#

use strict;
use Data::Dumper;
use Getopt::Std;
use File::Basename;

# prog's dir is part of @INC
use FindBin;
use lib "$FindBin::Bin";

our %cfg;
require 'iet-tools.conf';

my %opts;
getopts('hvc:t:l:d:p:',\%opts);

if ( defined($opts{'h'}) )
{
	usage();
	exit(0);
}

if ( ! -d $cfg{'state_dir'} )
{
	mkdir $cfg{'state_dir'};
}

process_command();
exit(0);

#-----------------------------------------------------------------------------------------------
# help
#-----------------------------------------------------------------------------------------------
sub usage
{
	print <<EOH;
$0 [-h] -c command

 state_raw		outputs the current state of /proc/net/iet
 state			outputs the current state of /proc/net/iet
 state_diff		determine if the current state matches the tracked state
			email operator report if changes occurred

 target_tid -t target	get target tid by name
 target_tid_new		get a new target tid [not thread-safe!]

 target_state -t target			output the state of target
 target_lun_path -t target -l lun	get a target's lun's path using target and lun #
 ddmap_lun_path -d ddmap			get a target's lun's path using /proc/net/iet/ddmap/
 lun_path_target -p lun_path		get a target using the lun's path

 target_sessions_remove -t target	disconnects all client sessions of a target
 target_initiators -t target 		list the initiator IP addresses of a target
 all_sessions_remove			disconnects all client sessions

 target_sessions	outputs current target session counters
 target_sessions_diff	target session listing showing current and saved (via stdin) session
			count per target
EOH
}

#-----------------------------------------------------------------------------------------------
# log err
#-----------------------------------------------------------------------------------------------
sub log_err
{
	my($msg) = @_;
	
	print "$msg\n";
}

#-----------------------------------------------------------------------------------------------
# logger
#-----------------------------------------------------------------------------------------------
sub logger
{
	my($msg) = @_;
	if ( $cfg{'verbose'} )
	{
		print "$msg\n";
	}
}

#-----------------------------------------------------------------------------------------------
#
# parse into a perl structure the contents of /proc/net/iet/entity
#
#tid:7 name:iqn.1990-01.edu.amherst.iscsi-target1:vm.rdm0
#        sid:10415673653920256 initiator:iqn.1998-01.com.vmware:vmware04-7baa6e84
#                cid:0 ip:148.85.141.159 state:active hd:none dd:none
#        sid:9571248723788288 initiator:iqn.1998-01.com.vmware:vmware03-49c0ad11
#                cid:0 ip:148.85.141.161 state:active hd:none dd:none
#-----------------------------------------------------------------------------------------------
sub proc_iet_dump_entity
{
	my ($targets, $proc_type) = @_;
	
	my $proc_file = "/proc/net/iet/$proc_type";
	
	logger("Processing file $proc_file");
	if ( ! -e $proc_file )
	{
		log_err("ERR> proc_iet_dump: $proc_file not found");
		return 0;
	}
	
	my $dest_proc_type = $proc_type;
	if ( $proc_type eq 'reservation' || $proc_type eq 'vpd')
	{
		$dest_proc_type = 'volume';
	}
	
	my $line;
	my $tid;
	my $target;
	my $index;
	open(FH, $proc_file) || die "unable to open $proc_file: $!";
	while(<FH>)
	{
		$line=$_;
		chomp($line);
		
		#
		# line starts with tid, we have a target
		#
		if ( $line =~ /^tid:(\d+) name:(.*)/ )
		{
			$tid = $1;
			$target = $2;
			$index = -1;
		}
		$$targets{$target}{'tid'} = $tid;
		
		#
		# line starts with a single tab, we have a new array entry for proc_type
		#
		if ( $line =~ /^\t\w+/ )
		{
			$index++;
		}
		
		#
		# vpd, fix spaces
		#
		if ( $line =~ /^\t\tvpd_/ )
		{
			$line =~ s/ //g;
		}
		
		#
		# line starts with some tab, make it part if the array entry
		# each array entry is a hash
		#
		# the arrays across volume, vpd and reservation have identical indexes
		#
		if ( $line =~ /^\t/ )
		{
			my @argval = split / /, $line;
			foreach my $item (@argval)
			{
				$item =~ s/\s+//g;
				my ($var, $val) = split /:/, $item, 2;
				if ( defined($var) && defined($val) )
				{
					$$targets{$target}{$dest_proc_type}[$index]{$var} = $val;
				}
				#print "item: $item  var=$var  val=$val\n";
			}
			
			#print Dumper \@argval;
			#print "NNN: $line\n";
		}
	}
	
	return 1;
}

#-----------------------------------------------------------------------------------------------
# fetch /etc/iet/initiators.allow and make data part of target
#-----------------------------------------------------------------------------------------------
sub etc_iet_initiators
{
	my ($targets, $type) = @_;
	
	my $initiators_file = "/etc/iet/initiators.$type";
	
	logger("Processing file $initiators_file");
	if ( ! -e $initiators_file )
	{
		log_err("ERR> etc_iet_initiators: $initiators_file not found");
		return 0;
	}
	
	my $line;
	my $target;
	open(FH, $initiators_file) || die "unable to open $initiators_file: $!";
	while(<FH>)
	{
		$line=$_;
		chomp($line);
		
		if ( $line =~ /^#/ )
		{
			next;
		}
		
		my($target, $ip_list) = split / /, $line;
		
		if ( defined($target) && defined($ip_list) && defined($$targets{$target}) )
		{
			my @list = split /,/, $ip_list;
			$$targets{$target}{"initiators_$type"} = \@list;
		}
	}
}

#-----------------------------------------------------------------------------------------------
# dump all of /proc/net/iet/* into a coherent perl structure
#-----------------------------------------------------------------------------------------------
sub proc_iet_dump
{
	my %targets;
	
	proc_iet_dump_entity(\%targets, 'session');
	proc_iet_dump_entity(\%targets, 'volume');
	proc_iet_dump_entity(\%targets, 'reservation');
	proc_iet_dump_entity(\%targets, 'vpd');
	etc_iet_initiators(\%targets, 'allow');
	return \%targets;
}

#-----------------------------------------------------------------------------------------------
# make a report for a given target
#-----------------------------------------------------------------------------------------------
sub target_dump
{
	my($target_name, $target) = @_;

	my($out, $vol_out, $session_out);

	my $vol_count = 0;
	foreach my $volume (@{$$target{'volume'}})
	{
		$vol_out .= ' lun '.
			$$volume{'lun'}.' '.
			$$volume{'path'}."\t".
			'['.join(' ',
				'type='.$$volume{'iotype'},
				'mode='.$$volume{'iomode'},
				'sn='.$$volume{'vpd_80_scsi_sn'},
				'reserved_by='.$$volume{'reserved_by'},
			)."]\n";
		$vol_count++;
	}

	my $session_count = 0;
	my @session_lines;
	foreach my $session (@{$$target{'session'}})
	{
		push @session_lines, '  '.$$session{'initiator'}."\t".
			'['.join(' ',
				's_ip='.$$session{'s_ip'},
				'd_ip='.$$session{'d_ip'},
				'cid='.$$session{'cid'},
				'sid='.$$session{'sid'},
			)."]";
		$session_count++;
	}
	$session_out = join("\n", sort @session_lines);

	$out = "$target_name [tid=$$target{'tid'} luns=$vol_count sessions=$session_count]\n";
	$out .= $vol_out.$session_out;
	
	my @lines = split /\n/, $out;
	my $line_no = 1;
	$out = "";
	foreach my $line (@lines)
	{
		$out .= sprintf("%02d %s\n",$line_no, $line);
		$line_no++;
	}
	
	return $out;
}

#-----------------------------------------------------------------------------------------------
# track iet state and write them out into the state dir
#-----------------------------------------------------------------------------------------------
sub state_track
{
	my($targets, $dir) = @_;
		
	foreach my $target (sort keys %{$targets})
	{
		my $file = "$dir/$target";
		logger("Writing target state $file");
		open(FH,">$file") || die "unable to open $file: $!";
		print FH target_dump($target, $$targets{$target});
		close FH;
	}
}

#-----------------------------------------------------------------------------------------------
# do a diff on the state in memory vs disk
#-----------------------------------------------------------------------------------------------
sub state_diff
{
	my($targets) = @_;

	
	my $dir_state = $cfg{'state_dir'};
	my $dir_tmp = $cfg{'state_tmp'};
	state_track($targets,$dir_tmp);
	my $report = "";
	foreach my $target (sort keys %{$targets})
	{
		#
		# optionally skip targets
		#
		my $skip = 0;
		foreach my $target_track_skip ( @{$cfg{'targets_track_skip'}} )
		{
			if ( $target =~ /$target_track_skip/ )
			{
				logger("skip: $target");
				$skip = 1;
			}
		}
		if ( $skip )
		{
			next;
		}
		
		if ( ! -e "$dir_state/$target" )
		{
			$report .= "new $dir_state/$target";
		}
		else
		{
			my $diff = `diff -c $dir_state/$target $dir_tmp/$target`;
			if ( $diff ne "" )
			{
				$report .= "----> $dir_state \n";
				$report .= `cat $dir_state/$target`;
				$report .= "\n----> $dir_tmp \n";
				$report .= `cat $dir_tmp/$target`;
				$report .= "\n----> diff \n";
				$report .= $diff;
				$report .= '-' x 90;
				$report .= "\n";
			}
		}
	}
	if ( length($report) == 0 )
	{
		logger('No changes found');
	}
	else
	{		
		logger('Changes found');
		open(MAIL, "| mail -s \"ietd /proc/net/iet diff on `hostname`\" $cfg{'operator'}")
			|| die "unable to spawn mail: $!";
		print MAIL $report;
		close MAIL;
		
		#
		# save the current state, so each time invoked we are
		# looking at the changes from one invocation to the next
		#
		state_track($targets,$cfg{'state_dir'});
	}
}

#-----------------------------------------------------------------------------------------------
# dump targets and each of their configuration and session state in human readable form
#-----------------------------------------------------------------------------------------------
sub state_dump
{
	my($targets) = @_;
	
	foreach my $target_name (sort keys %{$targets})
	{
		print target_dump($target_name, $$targets{$target_name});
		print '-' x 90;
		print "\n";
	}
}

#-----------------------------------------------------------------------------------------------
# target name -> tid
#-----------------------------------------------------------------------------------------------
sub target_tid
{
	my($targets, $target_name) = @_;

	if ( ! defined($$targets{$target_name}) )
	{
		return undef;
	}
	
	my $tid = $$targets{$target_name}{'tid'};
	return $tid;
}

#-----------------------------------------------------------------------------------------------
# next available tid
#-----------------------------------------------------------------------------------------------
sub target_tid_new
{
	my($targets) = @_;
	
	my $tid = 1;
	foreach my $target (keys %{$targets})
	{
		if ( $$targets{$target}{'tid'} > $tid )
		{
			$tid = $$targets{$target}{'tid'};
		}
	}
	$tid++;
	return $tid;
}

#-----------------------------------------------------------------------------------------------
# dump the state of a single target
#-----------------------------------------------------------------------------------------------
sub target_state
{
	my($targets, $target_name) = @_;
	
	if ( ! defined($$targets{$target_name}) )
	{
		return undef;
	}
	print target_dump($target_name, $$targets{$target_name});
}

#-----------------------------------------------------------------------------------------------
# get a target's lun's path (used to determine if a new LUN needs to be created or not)
#-----------------------------------------------------------------------------------------------
sub target_lun_path
{
	my($targets, $target_name, $lun_no) = @_;
	
	if ( ! defined($$targets{$target_name}) )
	{
		return undef;
	}
	
	my $volumes = $$targets{$target_name}{'volume'};
	my $path;
	foreach my $volume ( @{$volumes} )
	{
		if ( $$volume{'lun'} == $lun_no )
		{
			$path = $$volume{'path'};
			last;
		}
	}
	return $path;
}

#-----------------------------------------------------------------------------------------------
# get a target's lun's path using the ddmap filename
#-----------------------------------------------------------------------------------------------
sub ddmap_lun_path
{
	my($targets, $ddmap) = @_;
	
	logger("ddmap: $ddmap");
	
	my $basename = basename($ddmap);
	logger("basename: $basename");
	
	if ( $basename =~ /(\S*)\.(\d+)/ )
	{
		my $target_name = $1;
		my $lun_no = $2;
		
		logger("target_name: $target_name lun_no: $lun_no");
		return target_lun_path($targets, $target_name, $lun_no)
	}
	return undef;
}

#-----------------------------------------------------------------------------------------------
# search for a target given a lunpath - just to be clear, it is possible that
# a lunpath is in use by multiple targets? I would NOT recommend that - it will
# cause real problems if you mix fileio and blockio - just don't do that!
#-----------------------------------------------------------------------------------------------
sub lun_path_target
{
	my($targets, $lun_path) = @_;

	logger("l: $lun_path");
	foreach my $target (keys %{$targets})
	{
		foreach my $volume (@{$$targets{$target}{'volume'}})
		{
			if ( $$volume{'path'} eq $lun_path )
			{
				return "$target";
			}
		}
	}
	return undef;
}

#-----------------------------------------------------------------------------------------------
# clear all sessions given a target (if you don't block them also via iptables
# the clients will re-connect quickly again - got it?)
#-----------------------------------------------------------------------------------------------
sub target_sessions_remove
{
	my($targets, $target_name) = @_;

	if ( ! defined($$targets{$target_name}) )
	{
		log_err("ERR> unable to find $target_name");
		return undef;
	}
	
	my $tid = $$targets{$target_name}{'tid'};

	my $sessions = $$targets{$target_name}{'session'};
	my $out = "";
	foreach my $session ( @{$sessions} )
	{
		my $sid = $$session{'sid'};
		my $cid = $$session{'cid'};
		my $cmd = "ietadm --op delete --tid=$tid --sid=$sid --cid=$cid";
		$out .= "target_sessions_remove> $cmd (";
		$out .= `$cmd`;
		$out .= ")\n"
	}
	return $out;
}

#-----------------------------------------------------------------------------------------------
# clear all sessions, called during iscsi-target stop to ensure that
# ietadm --op delete will actually clear out all targets
#-----------------------------------------------------------------------------------------------
sub all_sessions_remove
{
	my($targets) = @_;

	my $out;
	foreach my $target_name (sort keys %{$targets})
	{
		#print Dumper \$$targets{$target_name};
		$out .= target_sessions_remove($targets, $target_name);

		#
		# clear out the luns
		#
		my $tid = $$targets{$target_name}{'tid'};
		foreach my $volume ( @{$$targets{$target_name}{'volume'}} )
		{
			my $lun = $$volume{'lun'};
			my $cmd = "ietadm --op delete --tid=$tid --lun=$lun";
			$out .= "all_sessions_remove> $cmd (";
			$out .= `$cmd`;
			$out .= ")\n"
		}
		
		#
		# clear out the target
		#
		my $cmd = "ietadm --op delete --tid=$tid";
		$out .= "all_sessions_remove> $cmd (";
		$out .= `$cmd`;
		$out .= ")\n"
		
	}
	return $out;
}

#-----------------------------------------------------------------------------------------------
# return target's initiator ip addresses
#-----------------------------------------------------------------------------------------------
sub target_initiators
{
	my($targets, $target_name, $block, $unblock) = @_;

	if ( ! defined($$targets{$target_name}) )
	{
		log_err("ERR> unable to find $target_name");
		return undef;
	}

	#
	# get a unique list of session initiator ip addresses
	#
	# this can be used by ietd_iptables start|stop
	#
	my %initiators;
	foreach my $session ( @{$$targets{$target_name}{'session'}} )
	{
		$initiators{$$session{'s_ip'}} = undef;
	}

	my $ip_list = join "\n", sort keys %initiators;
	
	return "$ip_list\n";
}

#-----------------------------------------------------------------------------------------------
# output current targets and their session count
#-----------------------------------------------------------------------------------------------
sub target_sessions
{
	my ($targets) = @_;
	
	foreach my $target (sort keys %{$targets})
	{
		my $session_count = 0;
		if ( defined($$targets{$target}{'session'}) )
		{
			$session_count = scalar(@{$$targets{$target}{'session'}});
		}
	
		print "$target\t$session_count\n";
	}
}

#-----------------------------------------------------------------------------------------------
# reading target<tab>session count from stdin, compare that to the current targets and session
# count values and present a report
#-----------------------------------------------------------------------------------------------
sub target_sessions_diff
{
	my ($targets) = @_;
	
	while(<STDIN>)
	{
		my $line = $_;
		
		my ($target, $session_count_max);
		if ( $line =~ /(.*)\t(.*)/ )
		{
			$target = $1;
			$session_count_max = $2;
		}
		else
		{
			print "line: $line mismatch\n";
			next;
		}
		
		my $session_count = 0;
		if ( defined($$targets{$target}{'session'}) )
		{
			$session_count = scalar(@{$$targets{$target}{'session'}});
		}
		
		my $status = "DIFF";
		if ( $session_count == $session_count_max )
		{
			$status = "OK";
		}
		
		printf("%4s %2d (expect: %2d) %s\n",$status, $session_count, $session_count_max, $target);
	}
}

#-----------------------------------------------------------------------------------------------
# the big switch...
#-----------------------------------------------------------------------------------------------
sub process_command
{
	if ( defined($opts{'v'}) )
	{
		$cfg{'verbose'} = 1;
	}
	
	if ( ! defined($opts{'c'}) )
	{
		usage();
		exit(1);
	}
	my $cmd = $opts{'c'};
	
	#
	# get current targets
	#
	my $targets = proc_iet_dump();

	#
	# command
	#
	if ( $cmd eq 'state_raw' )
	{
		$Data::Dumper::Varname = "iet_proc";
		print Dumper \$targets;
	}
	elsif ( $cmd eq 'state' )
	{
		state_dump($targets);
	}
	elsif ( $cmd eq 'state_diff' )
	{
		state_diff($targets);
	}
	elsif ( $cmd eq 'target_tid' )
	{
		my $tid = target_tid($targets, $opts{'t'});
		print "$tid\n";
	}
	elsif ( $cmd eq 'target_tid_new' )
	{
		my $tid = target_tid_new($targets);
		print "$tid\n";
	}
	elsif ( $cmd eq 'target_state' )
	{
		target_state($targets, $opts{'t'});
	}
	elsif ( $cmd eq 'target_lun_path' )
	{
		my $path = target_lun_path($targets, $opts{'t'}, $opts{'l'});
		print "$path\n";
	}
	elsif ( $cmd eq 'ddmap_lun_path' )
	{
		my $path = ddmap_lun_path($targets, $opts{'d'});
		print "$path\n";
	}
	elsif ( $cmd eq 'lun_path_target' )
	{
		my $target = lun_path_target($targets, $opts{'p'});
		print "$target\n";
	}
	elsif ( $cmd eq 'target_sessions_remove' )
	{
		my $result = target_sessions_remove($targets, $opts{'t'});
		if ( ! defined($result) )
		{
			exit(1);
		}
		print $result;
	}
	elsif ( $cmd eq 'all_sessions_remove' )
	{
		my $result = all_sessions_remove($targets);
		if ( !defined($result) )
		{
			exit(1);
		}
		print $result;
	}
	elsif ( $cmd eq 'target_initiators' )
	{
		my $result = target_initiators($targets, $opts{'t'});
		if ( ! defined($result) )
		{
			exit(1);
		}
		print $result;
	}
	elsif ( $cmd eq 'target_sessions' )
	{
		target_sessions($targets);
	}
	elsif ( $cmd eq 'target_sessions_diff' )
	{
		target_sessions_diff($targets);
	}
	else
	{
		log_err('ERR> invalid command');
		exit(1);
	}
}
exit(0);
