#!/bin/sh

#
# ddmap control script (start, shutdown, query)
#
# ddmap is a feature of iet that provides you with a
# bitmap of block changes. Each bit encodes 16KB.
# We use this bitmap to determine what segments of
# disk to read when processing a LVM snapshot.
#
# on a per lun basis do:
#
# on shutdown if dirty keep it dirty or if active save ddmap and set clean
#
# on start if active, then we have a dirty state
# on start if dirty, then we do nothing (keep dirty)
#
# dirty state is cleared once a full backup was done
#
# Steffen Plotner, 2010
#

. /etc/iscsi-target/ietd_config.sh

#
# ddmap states:
#	active->clean
#	active->dirty->clean
#

#------------------------------------------------------------------------------
# awaken a ddmap file: this does nothing special other than to manage the state
# of the ddmap file, so if you find that the state was active, then state
# automatically becomes dirty (i.e. we had a shutdown state that left the map
# active (which means the shutdown did not work), hence the map becomes dirty)
#------------------------------------------------------------------------------
function ddmap_awake()
{
	local ddmap="$1"

	last_state=$(ddmap_state_get $ddmap)
	logmsg "ddmap> $ddmap last state: $last_state (awake)"

	case "$last_state" in
		#
		# after a system crash, the active state would be left behind
		# hence we now switch to a dirty state
		#
		DDMAP_ACTIVE|DDMAP_UNKNOWN)
			logmsg "ddmap> setting state to dirty since it was last active (unclean shutdown)"
			ddmap_state_set "$ddmap" "DDMAP_DIRTY"
			;;
		#
		# after a proper shutdown, we switch from clean to active
		#
		DDMAP_CLEAN)
			logmsg "ddmap> setting state to active"
			ddmap_state_set "$ddmap" "DDMAP_ACTIVE"
			;;
		#
		# if it is already dirty, just keep it that way
		#
		DDMAP_DIRTY)
			logmsg "ddmap> keeping state $last_state"
			;;
		*)
			logmsg "ddmap_err> unkown state $last_state"
			echo_failure
			echo
			exit 1
	esac
}

#------------------------------------------------------------------------------
# put a ddmap to sleep (i.e. get the ddmap from /proc into the current file)
#------------------------------------------------------------------------------
function ddmap_sleep()
{
	local ddmap="$1"
	local action="$2"
	local proc_ddmap="$3"
	# used by ddmap_combine
	local ddmap_copy_dir="$4"
	
	local ccode
	local cmd
	local out

	last_state=$(ddmap_state_get $ddmap)
	logmsg "ddmap> $ddmap last state: $last_state (action: $action)"

	case "$action" in
		#
		# status
		#
		status)
			echo "${ddmap}.ddmap $last_state"
			;;

		#
		# stop or ddmap_dump, the dumping assumes that the map was previously suspended(!)
		#
		stop|ddmap_dump)

			case "$last_state" in
				DDMAP_ACTIVE)
					#
					# for time reasons, we have to just dump the map content quickly,
					# we will merge the contents dumped at the time when the snapshot
					# is actually taken, we need the stop event to get done fast!
					#
					local ts=$(date +%Y%m%d.%H%M%S.%N)

					cmd="dd if=$proc_ddmap of=$VGMETA_DIR/$ddmap.$ts.ddmapdump bs=1k"
					logmsg "ddmap> cmd: $cmd"

					out=$($cmd 2>&1)
					logmsg "ddmap> out: $out"

					ddmap_log "$ddmap" "dumped to $ddmap.$ts.ddmapdump"

					#
					# we have dumped and stored the state of the map, we are now
					# clean for the next startup or next snapshot, the ddmap_combine
					# action will set the state back to active of course!
					#
					ddmap_state_set "$ddmap" "DDMAP_CLEAN"
					;;
				DDMAP_UNKNOWN)
					ddmap_state_set "$ddmap" "DDMAP_DIRTY"
					;;
				DDMAP_DIRTY|DDMAP_CLEAN)
					logmsg "ddmap> keeping state $last_state"
					;;
				*)
					logmsg "ddmap_err> unkown state $last_state"
					echo_failure
					echo
					exit 1
			esac
			;;
		
		#
		# combine ddmap maps
		#
		ddmap_combine)
			
			#
			# combine all .ddmapdump files into a single ddmap file
			#
			if [ -e "$VGMETA_DIR/$ddmap.ddmap" ]; then
				rm "$VGMETA_DIR/$ddmap.ddmap"
			fi
			
			for ddmapdump in $VGMETA_DIR/$ddmap*.ddmapdump; do
				#
				# the wildcard above can cause problems if there are no
				# files, it returns the '*' in $proc_ddmap - so validate it
				#
				if [ ! -e $ddmapdump ]; then
					continue
				fi
				
				echo "ddmap> ddmapdump: $ddmapdump"
				
				cmd="ddmap -v -a -s $ddmapdump -t $VGMETA_DIR/$ddmap.ddmap"
				echo "ddmap> cmd: $cmd"
				out=$($cmd)
				echo "$out"
				ddmap_log "$ddmap" "merged $ddmapdump"
				rm $ddmapdump
			done

			#
			# copy resulting ddmap file into user supplied
			# space and include the current state also
			#
			# so, this is important, we take the current state
			# here and copy it to another location, the reader
			# of that data has to ensure to do the right thing.
			# If the state is dirty, read the whole disk...
			#
			logmsg "ddmap> copy ddmap to $ddmap_copy_dir"
			if [ -d $ddmap_copy_dir ]; then
				#
				# the data file has to be moved out of the
				# way, since the above ddmap -s -t merges data(!)
				#
				# we might not have a map file, if the volume was
				# new (hence dirty state)
				#
				if [ -e "$VGMETA_DIR/$ddmap.ddmap" ]; then
					mv -v "$VGMETA_DIR/$ddmap.ddmap" $ddmap_copy_dir
				fi
				cp -v "$VGMETA_DIR/$ddmap.state" $ddmap_copy_dir
			fi
			
			#
			# set the state to clean and then active again
			#
			ddmap_state_set "$ddmap" "DDMAP_CLEAN"
			ddmap_state_set "$ddmap" "DDMAP_ACTIVE"
			;;
	esac
}

#------------------------------------------------------------------------------
# start
#------------------------------------------------------------------------------
function start()
{
	echo -n "Starting iSCSI Target: ddmap"
	
	#
	# enumerate each ddmap provided by ddmap_dir, notice that we go after the
	# state file, since there is no .ddmap file anymore.. Notice the use of
	# find -wholename it matches .state files under ddmap directories only.
	#
	for ddmap in $(cd $VGMETA_DIR; find -wholename "*/ddmap/*.state" -printf '%P\n'); do
		#
		# strip extension
		#
		ddmap=$(echo "$ddmap" | sed -e 's/\.state$//')

		ddmap_awake "$ddmap"
	done
	echo_success
	echo
}

#------------------------------------------------------------------------------
# stop
#------------------------------------------------------------------------------
function stop()
{
	local action="$1"
	local lun_path_match="$2"
	local ddmap_copy_dir="$3"

	if [ $action = "stop" ]; then
		echo -n "Stopping iSCSI Target: ddmap"
	fi

	#
	# iet must be running in order to query /proc/net/iet/ddmap
	#
	# calling code is responsible for blocking iscsi traffic to
	# ensure that the map we read here is steady
	#
	if [ ! -d "$DDMAP_PROC" ]; then
		logmsg "ddmap_err> on stop expecting $DDMAP_PROC"
		echo_failure
		echo
		exit 1
	fi
	
	#
	# enumerate each ddmap provided by iet
	#
	# iet provides the maps using target.lun notation.
	# Using iet-tools.plx we convert the ddmap reference to a
	# backing store reference which could be a file or a LVM volume.
	#
	for proc_ddmap in $DDMAP_PROC/*; do
		#
		# the wildcard above can cause problems if there are no
		# files, it returns the '*' in $proc_ddmap - so validate it
		#
		if [ ! -e $proc_ddmap ]; then
			continue
		fi
		
		#
		# convert that to a file path, either block device or plain file
		#
		lun_path=$($ROOT/iet-tools.plx -c ddmap_lun_path -d $proc_ddmap)

		#
		# we process all ddmap entries when stopping and obtaining the status
		#
		case "$action" in
			status|stop)
				true
				;;
			*)
				if [ "$lun_path" != "$lun_path_match" ]; then
					continue
				fi
				;;
		esac

		logmsg "ddmap> ddmap: $proc_ddmap"
		logmsg "ddmap> lun_path: $lun_path (action: $action)"

		ddmap=""
		ddmap_dir=""
		if [ -b "$lun_path" ]; then
			#
			# lvm block device (expected 3 arguments, /dev/vg/lv)
			#
			dev=$(echo "$lun_path" | cut -f2 -d/)
			vg=$(echo "$lun_path" | cut -f3 -d/)
			lv=$(echo "$lun_path" | cut -f4 -d/)
			if [ "$dev" != "dev" -o -z "$vg" -o -z "$lv" ]; then
				logmsg "ddmap_err> unsupported block device name, got $lun_path, requires /dev/vg/lv"
				continue
			fi
			ddmap_dir="$VGMETA_DIR/$vg/ddmap"
			ddmap="$vg/ddmap/$lv"
		else
			#
			# unknown
			#
			logmsg "ddmap_err> unknown lun_path: $lun_path from $proc_ddmap, skipping (must be a block device)"
			continue
		fi

		#
		# ensure ddmap dir exists
		#
		mkdir -p "$ddmap_dir"

		#
		# handle the actions of this ddmap
		#
		case "$action" in
			mapsuspend)
				echo "ddmap> mapsuspend $proc_ddmap"
				echo -n 'mapsuspend' > $proc_ddmap
				ddmap_log "$ddmap" 'mapsuspend'
				#ddmap -a -s $proc_ddmap
				;;
			mapresume)
				echo "ddmap> mapresume $proc_ddmap"
				echo -n 'mapresume' > $proc_ddmap
				ddmap_log "$ddmap" 'mapresume'
				#ddmap -a -s $proc_ddmap
				;;
			mapzap)
				echo "ddmap> mapzap $proc_ddmap"
				echo -n 'mapzap' > $proc_ddmap
				ddmap_log "$ddmap" 'mapzap'
				;;
			status|stop|ddmap_dump|ddmap_combine)
				ddmap_sleep "$ddmap" "$action" "$proc_ddmap" "$ddmap_copy_dir"
				;;
		esac

	done
	
	if [ $action = "stop" ]; then
		echo_success
		echo
	fi
}

cmd="$1"
shift
case "$cmd" in
	start_internal)
		#
		# processes all ddmaps
		#
		start
		;;
	stop_internal)
		#
		# processes all targets/luns
		#
		stop 'stop'
		;;
	ddmap_dump)
		stop 'ddmap_dump' $*
		;;
	ddmap_combine)
		stop 'ddmap_combine' $*
		;;
	mapsuspend)
		stop 'mapsuspend' $*
		;;
	mapresume)
		stop 'mapresume' $*
		;;
	mapzap)
		stop 'mapzap' $*
		;;
	status)
		stop 'status'
		;;
	*)
		echo $"Usage: $0 {status}"
		exit 1
esac
exit 0
