#!/bin/sh

#
# iscsi-target configuration update (multihost aware)
#
# Steffen Plotner, 2005-2010
#

. /etc/iscsi-target/ietd_config.sh
. /etc/iscsi-target/targets.conf

#------------------------------------------------------------------------------
# iet online target
#------------------------------------------------------------------------------
function target_online
{
	local tgt_name="$1"
	local tid
	local cmd
	local ccode
	
	if [ -z "$tgt_name" ]; then
		echo "tgt_err> tgt_name parameter is missing"
		exit 1
	fi
	
	#
	# check if there is a tid for this target
	#
	tid=$(target_tid "$tgt_name")
	ccode=$?
	if [ $ccode -eq 1 ]; then
		echo "tgt_err> failed to get target_tid from target_name: $tgt_name"
		exit 1
	fi

	if [ -z "$tid" ]; then
		tid=$($ROOT/iet-tools.plx -c target_tid_new)
		echo "tgt> tid: $tid (new)"

		cmd="ietadm --op new --tid=$tid --params Name=$tgt_name"
		echo "tgt> $cmd"
		$cmd
		ccode=$?
		if [ $ccode != 0 ]; then
			echo "tgt_err> ietadm returned error: $ccode"
			exit 1
		fi
	fi
}

#------------------------------------------------------------------------------
# target offline: disconnects iscsi initiator sessions
#------------------------------------------------------------------------------
function target_offline
{
	local tgt_name="$1"
	local lun_id="$2"
	local lun_path="$3"
	local vg="$4"

	#
	# get a list of active session source ip addresses
	# block tcp stream for the target's session source ip addresses
	#
	local initiators=$($ROOT/iet-tools.plx -c target_initiators -t "$tgt_name")
	local initiator
	for initiator in $initiators; do
		echo "tgt> initiator: $initiator (blocking)"
		$ROOT/ietd_iptables.sh stop $initiator
	done

	#
	#
	# drop all active sessions from this target
	#
	$ROOT/iet-tools.plx -c target_sessions_remove -t "$tgt_name"

	#
	# remove lun from ietd
	#
	cmd="ietadm --op delete --tid=$current_tid --lun=$lun_id"
	echo "tgt> $cmd"
	$cmd
	ccode=$?
	if [ $ccode != 0 ]; then
		echo "tgt_err> ietadm returned error code $ccode"
		exit 1
	fi
	sleep 2

	#
	# remove target from ietd
	#
	cmd="ietadm --op delete --tid=$current_tid"
	echo "tgt> $cmd"
	$cmd
	ccode=$?
	if [ $ccode != 0 ]; then
		echo "tgt_err> ietadm returned error code $ccode"
		exit 1
	fi
	sleep 2

	#
	# unblock tcp stream
	#
	for initiator in $initiators; do
		echo "tgt> initiator: $initiator (unblocking)"
		$ROOT/ietd_iptables.sh start $initiator
	done
}

#------------------------------------------------------------------------------
# iet lun online
#------------------------------------------------------------------------------
function lun_online
{
	local vg="$1"
	local tgt_name="$2"
	local lun_id="$3"
	local lun_path="$4"
	local lun_parms="$5"

	local cmd
	local ccode

	#
	# check lun path (only in online mode - when offline, it is not clear if the
	# paths are valid)
	#
	if [ ! -e $lun_path ]; then
		echo "tgt_err> lun path: $lun_path not found."
		exit 1
	fi
	
	#
	# check if lun number is already assigned, if so, do nothing, else
	# create new lun.
	#
	local ietd_lun_path=$($ROOT/iet-tools.plx -c target_lun_path -t "$tgt_name" -l "$lun_id")
	local lun_size
	if [ -z "$ietd_lun_path" ]; then
		#
		# create new lun
		#
		cmd="ietadm --op new --tid=$current_tid --lun=$lun_id --params=Path=$lun_path,$lun_parms"
		echo "tgt> $cmd"
		$cmd
		ccode=$?
		if [ $ccode != 0 ]; then
			echo "tgt_err> ietadm returned error: $ccode"
			exit 1
		fi
		
		#
		# track the lunsize (lvm only)
		#
		lun_size=$(lv_size_get "$lun_path")
		lv_track_lun_size_set "$vg" "$lun_path" "$lun_size"
		echo "tgt> tracked lun size: $lun_size"
	else
		#
		# verify that the path agrees between the config and active system
		#
		# this test in theory is not needed anymore, since we have decided to
		# go to a single lun per target layout - however it would be good to 
		# make sure things are correct!
		#
		if [ "$lun_path" != "$ietd_lun_path" ]; then
			echo "tgt_err> lun path mismatch, you must remove the active LUN manually using ietadm"
			echo "tgt_err> lun_path: $lun_path"
			echo "tgt_err> ietd_lun_path: $ietd_lun_path"
			exit 1
		fi

		#
		# obtain the tracked volume size (if there is none, then track it now)
		# if there is one, compare it to the current size and if that is different
		# we need to do a ietd fixup to have it recognize the new volume size
		#
		lun_size=$(lv_size_get "$lun_path")
		echo "tgt> lun: $lun_id, path: $ietd_lun_path lun size: $lun_size"
		local tracked_lun_size=$(track_lun_size_get "$vg" "$lun_path")
		if [ -z "$tracked_lun_size" ]; then
			echo "tgt> tracked lun size not found, lun is $lun_size, tracking it now (OK)"
			lv_track_lun_size_set "$vg" "$lun_path" "$lun_size"
		else
			#
			# check of the lun_size is different from the tracked_lun_size
			#
			if  [ "$lun_size" = "$tracked_lun_size" ]; then
				return
			fi
			echo "tgt> lun size was $tracked_lun_size, current $lun_size (UPDATING)"

			echo "--LUN RESIZE UPDATE (BEGIN)"
			
			#
			# get a list of active session source ip addresses
			# block tcp stream for the target's session source ip addresses
			#
			local initiators=$($ROOT/iet-tools.plx -c target_initiators -t "$tgt_name")
			local initiator
			for initiator in $initiators; do
				echo "tgt> initiator: $initiator (blocking)"
				$ROOT/ietd_iptables.sh stop $initiator
			done

			#
			#
			# drop all active sessions from this target
			#
			$ROOT/iet-tools.plx -c target_sessions_remove -t "$tgt_name"

			#
			# remove lun from ietd
			#
			cmd="ietadm --op delete --tid=$current_tid --lun=$lun_id"
			echo "tgt> $cmd"
			$cmd
			ccode=$?
			if [ $ccode != 0 ]; then
				echo "tgt_err> ietadm returned error code $ccode"
				exit 1
			fi
			sleep 2

			#
			# add lun to ietd
			#
			cmd="ietadm --op new --tid=$current_tid --lun=$lun_id --params=Path=$lun_path,$lun_parms"
			echo "tgt> $cmd"
			$cmd
			ccode=$?
			if [ $ccode != 0 ]; then
				echo "tgt_err> ietadm returned error code $ccode"
				exit 1
			fi
			sleep 2

			#
			# unblock tcp stream (this will cause re-connections to target)
			#
			for initiator in $initiators; do
				echo "tgt> initiator: $initiator (unblocking)"
				$ROOT/ietd_iptables.sh start $initiator
			done

			#
			# track the size of the lun (LVM only supported)
			#
			lun_size=$(lv_size_get "$lun_path")
			lv_track_lun_size_set "$vg" "$lun_path" "$lun_size"

			echo "--LUN RESIZE UPDATE (END)"
		fi
	fi
}

#------------------------------------------------------------------------------
# targets, only configure targets for which the volume group is active on
# the current host
#------------------------------------------------------------------------------
function target
{
	local tgt_state="$1"
	local lun_path="$2"
	local iotype="$3"
	local scsisn="$4"
	local pg="$5"
	local ig="$6"

	if [ -z "$tgt_state" -o -z "$lun_path" -o -z "$iotype" -o -z "$scsisn" -o -z "$pg" -o -z "$ig" ]; then
		echo "tgt_err> target up|down lun_path iotype scsisn portalgroup initiatorgroup"
		echo "tgt_err> target $tgt_state $lun_path $iotype $scsisn $pg $ig"
		exit 1
	fi

	if [ "$tgt_state" != "up" -a "$tgt_state" != "down" ]; then
		echo "tgt_err> invalid target state: $tgt_state, expecting up or down"
		exit 1
	fi

	#
	# lvm block device (expected 3 arguments, /dev/vg/lv)
	#
	local dev=$(echo "$lun_path" | cut -f2 -d/)
	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	if [ "$dev" != "dev" -o -z "$vg" -o -z "$lv" ]; then
		echo "tgt_err> unsupported block device name, got $lun_path, requires /dev/vg/lv"
		return
	fi
	
	#
	# skip this target definition if the volume group is not active on this host
	#
	target_valid=1
	if [ ! -d "/dev/$vg" ]; then
		target_valid=0
		return
	fi

	#
	# target callback
	#
	conf_target_callback "$dev" "$vg" "$lv"
	
	#
	# compute target name (iqn do not allow the underscore)
	#
	local tgt_name="$IET_TARGET_PREFIX.$pg.$vg.$lv"
	tgt_name=$(echo "$tgt_name" | sed -e 's/_/-/g')
	
	echo "tgt> tgt_name: $tgt_name ($tgt_state)"
	
	#
	# notice that the conf variables are NOT local, this allows
	# the option function to still work within a target
	#
	iet_conf="$VGMETA_DIR/$vg/iet/ietd.conf"
	iet_portal_conf="$VGMETA_DIR/$vg/iet/ietd_portal.conf"
	ini_allow="$VGMETA_DIR/$vg/iet/initiators.allow"
	tgt_allow="$VGMETA_DIR/$vg/iet/targets.allow"

	local lun_id="0"

	#
	# configure target
	#
	if [ $run_mode = "ietd_online" ]; then
		#
		# either create or delete target
		#
		if [ "$tgt_state" = "up" ]; then
			#
			# in online mode we also remember the current_tid
			# as it will be used by the options statements
			#
			target_online "$tgt_name"
			current_tid=$(target_tid "$tgt_name")
		else
			#
			# destroy the target, this does NOT delete any data!
			#
			current_tid=$(target_tid "$tgt_name")
			if [ -n "$current_tid" ]; then
				target_offline "$tgt_name" "$lun_id" "$lun_path" "$vg"
			else
				echo "tgt> $tgt_name was already down"
			fi
		fi
	fi
	
	#
	# targets that are turned off, no more conf for this target and subsequent
	# statements.
	#
	if [ "$tgt_state" = "down" ]; then
		target_valid=0
		return
	fi
	
	echo "" >> $iet_conf
	echo "Target $tgt_name" >> $iet_conf

	echo "" >> $iet_portal_conf
	echo "Target $tgt_name" >> $iet_portal_conf
	echo " RedirectExec IET_PORTAL_REDIRECT_EXEC" >> $iet_portal_conf
	
	#
	# configure target options that are common to all targets
	#
	conf_target_options
	
	#
	# Lun 0 (only!)
	#
	local lun_parms="Type=$iotype,ScsiId=$scsisn,ScsiSN=$scsisn"
	
	if [ $run_mode = "ietd_online" ]; then
		lun_online "$vg" "$tgt_name" $lun_id "$lun_path" "$lun_parms"
	fi
	echo " Lun $lun_id Path=$lun_path,$lun_parms" >> $iet_conf
	
	#
	# configure initiators.allow
	#
	#local tg_init=$(gen_ini_allow_entry "$tgt_name" "iqn.1986-03.com.sun:01:e00000000000.4c9de97f $ig")
	local tg_init=$(gen_ini_allow_entry "$tgt_name" "$ig")
	echo "$tg_init" >> $ini_allow
	
	#
	# configure targets.allow
	#
	# they cannot be configured until the services start, because
	# the configuration is host specific. I.e. if you refer to target group
	# vlan141 on node1 it would mean 141.70,141.71. Then on node2 it is 141.72,141.73
	# If you pregenerate the resulting IP addresses and the volume group moves
	# to a new node, then the targets.allow will be wrong.
	#
	# the other way to handle this is to configure targets.allow with all
	# known ip address - it would work, except that here in this code path
	# we are running on the local host.
	#
	echo "$tgt_name $pg" >> $tgt_allow
}

#------------------------------------------------------------------------------
# configure options specific to target or global options
#
# re-specifying an option within a target is allowed and will override the
# global target setting. The last option counts for ietd.
#------------------------------------------------------------------------------
function option
{
	local op_name="$1"
	local op_val="$2"
	local cmd
	local ccode

	if [ $target_valid -eq 0 ]; then
		return
	fi
	
	if [ -z "$op_name" -o -z "$op_val" ]; then
		echo "tgt_err> option name value"
		echo "tgt_err> option $op_name $op_val"
		exit 1
	fi

	#
	# global option mode (there is no online mode for this!)
	#
	if [ $option_mode = "global" ]; then
		echo "$op_name $op_val" >> $iet_conf
		return
	fi

	#
	# target option mode
	#
	if [ $run_mode = "ietd_online" ]; then
		cmd="ietadm --op update --tid=$current_tid --params=$op_name=$op_val"
		$cmd
		ccode=$?
		if [ $ccode != 0 ]; then
			echo "tgt_err> $cmd"
			echo "tgt_err> ietadm returned error: $ccode"
			exit 1
		fi
	fi
	echo " $op_name $op_val" >> $iet_conf
}

#------------------------------------------------------------------------------
# CHAP incoming/outgoing user accounts
#------------------------------------------------------------------------------
function user
{
	local relation="$1"
	local username="$2"
	local secret="$3"

	local cmd_pre
	local cmd
	local ccode
	local cfg

	if [ $target_valid -eq 0 ]; then
		return
	fi
	
	if [ -z "$relation" -o -z "$username" -o -z "$secret" ]; then
		echo "tgt_err> user in|out username secret"
		echo "tgt_err> you can have many in coming usernames and only ONE outgoing entry"
		exit 1
	fi

	if [ "$relation" = "out" ]; then
		relation="OutgoingUser"
	else
		relation="IncomingUser"
	fi

	#
	# global usernames
	#
	if [ $option_mode = "global" ]; then
		cfg="$relation $username $secret"
		cmd_pre="ietadm --op delete --user --params=$relation=$username"
		cmd="ietadm --op new --user --params=$relation=$username,Password=$secret"
	else
		#
		# target specific usernames (send those to the portal also, some qla4010 needs it)
		#
		cfg=" $relation $username $secret"
		cmd_pre="ietadm --op delete --user --tid=$current_tid --params=$relation=$username"
		cmd="ietadm --op new --tid=$current_tid --user --params=$relation=$username,Password=$secret"
		echo "$cfg" >> $iet_portal_conf
	fi

	echo "$cfg" >> $iet_conf
	if [ $run_mode = "ietd_online" ]; then
		$cmd_pre 2> /dev/null
		$cmd
		ccode=$?
		if [ $ccode != 0 ]; then
			echo "tgt_err> $cmd"
			echo "tgt_err> ietadm returned error: $ccode"
			exit 1
		fi
	fi
}

#------------------------------------------------------------------------------
# configure targets on each host
#------------------------------------------------------------------------------
function configure_targets
{
	local host_current=$(uname -n)
	local host
	for host in $IET_TARGET_HOSTS; do
		echo "tgt> host: $host"

		if [ $host_current != $host ]; then
			host_wrapper="ssh $host"
		fi
		$host_wrapper /etc/iscsi-target/ietd_targets.sh configure_local $*
	done
}

#------------------------------------------------------------------------------
# configure targets on current host
#------------------------------------------------------------------------------
function configure_localhost_targets
{
	#
	# configure run mode, online or offline
	#
	if [ -z "$1" -o "$1" = "online" ]; then
		run_mode="ietd_online"
		echo "tgt> run mode: online"
		if [ ! -e /proc/net/iet/volume ]; then
			echo "tgt> /proc/net/iet/volume not found, use '$0 configure offline' instead"
			unlock
			exit 0
		fi
	else
		echo "tgt> run mode: offline"
	fi
	
	#
	# prepare iet dir in volume groups that are active on this host
	#
	local vg
	local vgs=$(vg_active_get $host)
	for vg in $vgs; do
		#
		# iscsi-target configuration
		#
		mkdir -p "$VGMETA_DIR/$vg/iet"
		
		local info=$(cat <<EOF
#------------------------------------------------------------------------------
# Generated on: `date`
# Generated by: $host
# Volume Group: $vg
#------------------------------------------------------------------------------
)
		echo "$info" > "$VGMETA_DIR/$vg/iet/ietd.conf"
		echo "$info" > "$VGMETA_DIR/$vg/iet/ietd_portal.conf"
		echo "$info" > "$VGMETA_DIR/$vg/iet/initiators.allow"
		echo "$info" > "$VGMETA_DIR/$vg/iet/targets.allow"
		
		echo "# currently not used!!!!!" >> "$VGMETA_DIR/$vg/iet/ietd_portal.conf"
		
		#
		# track lunsize changes
		#
		mkdir -p "$VGMETA_DIR/$vg/lunsize"

		#
		# target pre callback (this is volume group specific)
		#
		# must be done before targets.d (dependency)
		#
		conf_target_callback_pre "$vg"

		#
		# allow programmable target definitions (yes, they
		# must be placed on the host that currently holds the
		# resource) and they consist of target.d snippets, just
		# like your target.conf file
		#
		mkdir -p "$VGMETA_DIR/$vg/targets.d"
		local tfile
		for tfile in $(cd $VGMETA_DIR/$vg/targets.d; find -name "*.conf" -printf '%P\n'); do
			. "$VGMETA_DIR/$vg/targets.d/$tfile"
		done
	done
	
	#
	# configure targets that are part of this host
	# (target.conf)
	#
	conf_targets
	
	#
	# enumerate all volume groups, send ietd, initiators.allow and targets.allow
	# to each of the portals
	#
	# the ietd-portal has been disabled, because ietd itself is single threaded,
	# and on failover that is not good because all initiators ask ietd at the
	# portal at the same time for redirection information. We have replaced that
	# with the iscsi-portal-daemon code until we have time to re-work the ietd.
	#
	#local portal
	#for vg in $vgs; do
	#	echo "tgt> sending volume group: $vg config to iscsi-portal cluster..."
	#	for portal in $IET_PORTAL_HOSTS; do
	#		scp -q "$VGMETA_DIR/$vg/iet/ietd_portal.conf" "$portal:/etc/iscsi-target/data/${vg}_ietd_portal.conf"
	#		scp -q "$VGMETA_DIR/$vg/iet/initiators.allow" "$portal:/etc/iscsi-target/data/${vg}_initiators.allow"
	#		scp -q "$VGMETA_DIR/$vg/iet/targets.allow" "$portal:/etc/iscsi-target/data/${vg}_targets.allow"
	#	done
	#done
	
	#
	# reload the portal state (service restart is wrong, no locks!)
	#
	#for portal in $IET_PORTAL_HOSTS; do
	#	ssh $portal /etc/iscsi-target/ietd_portal.sh reload
	#done
}

#------------------------------------------------------------------------------
# configure the localhost iet using mounted volume group configurations
# and resolving the targets.allow portal groups
#------------------------------------------------------------------------------
function configure_localhost_iet
{
	echo -n "Starting iSCSI Target: targets"

	#
	# write config headers
	#
	local info=$(cat <<EOF
#------------------------------------------------------------------------------
# iSCSI Enterprise Target Cluster Configuration
#
# The configuration, cluster and ddmap features are designed
# by Steffen Plotner, 2010
#
# Generated on: `date`
# Generated by: $host
#------------------------------------------------------------------------------
# Do NOT modify this file, it is maintained by /etc/iscis-target/targets.conf.
# Each volume group maintains iscsi-target configuration files which are
# merged at startup time of the iscsi-target based on the active volume groups
# of a host.
#------------------------------------------------------------------------------
)
	echo "$info" > "$IET_CONF_DIR/ietd.conf"
	echo "$info" > "$IET_CONF_DIR/initiators.allow"
	echo "$info" > "$IET_CONF_DIR/targets.allow"
	> "$IET_CONF_DIR/targets.temp"
	logmsg "iet> configuration startup on host $host"

	#
	# configure global options (must enable target valid, other user or
	# options commands get skipped)
	#
	iet_conf="$IET_CONF_DIR/ietd.conf"
	target_valid=1
	conf_global
	
	#
	# for active volume groups, gather ietd.conf and initiators.allow
	# get the targets.allow into a temp file processed below
	#
	local vg
	local vgs=$(vg_active_get $host)
	for vg in $vgs; do
		logmsg "iet> active volume group: $vg"
		cat "$VGMETA_DIR/$vg/iet/ietd.conf" >> "$IET_CONF_DIR/ietd.conf"
		cat "$VGMETA_DIR/$vg/iet/initiators.allow" >> "$IET_CONF_DIR/initiators.allow"
		cat "$VGMETA_DIR/$vg/iet/targets.allow" >> "$IET_CONF_DIR/targets.temp"
	done

	#
	# generate portal group definitions (enum interfaces and gather common portal group names
	# and their associated IP addresses)
	#
	rm $IET_CONF_DIR/*.portal_group 2> /dev/null
	local ifname
	for ifname in /etc/sysconfig/network-scripts/ifcfg-eth*; do
		#
		# include interface config, for those that have a custom
		# route table name configure them
		#
		IET_PORTAL_GROUP=""
		. $ifname
		if [ -n "$IET_PORTAL_GROUP" ]; then
			logmsg "iet> $ifname has portal group $IET_PORTAL_GROUP with ip $IPADDR"
			echo "$IPADDR" >> $IET_CONF_DIR/$IET_PORTAL_GROUP.portal_group
		fi
	done

	#
	# generate targets.allow from targets.temp
	#
	local portal_group
	for portal_group in $(cd $IET_CONF_DIR; find -name "*.portal_group" -printf '%P\n'); do
		#
		# strip extension
		#
		portal_group=$(echo "$portal_group" | sed -e 's/\.portal_group$//')
		logmsg "iet> portal group name: $portal_group"
		
		#
		# produce a comma separated list of ip addresses for this portal
		#
		local tg_pg
		local ip
		tg_pg=""
		for ip in $(cat "$IET_CONF_DIR/$portal_group.portal_group"); do
			tg_pg="${tg_pg}${ip},"
		done
		tg_pg=$(echo "$tg_pg" | sed -e 's/,$//')
		logmsg "iet> portal group ip: $tg_pg"

		#
		# foreach given portal group, find the entries in the targets.temp
		# file and translate them into the current system ip addresses
		#
		cat "$IET_CONF_DIR/targets.temp" | \
			grep "$portal_group$" | \
			sed -e "s/ $portal_group$/ $tg_pg/" >> "$IET_CONF_DIR/targets.allow"
	done

	echo_success
	echo
}

#------------------------------------------------------------------------------
# generate initiator.allow line
#------------------------------------------------------------------------------
function gen_ini_allow_entry
{
	local tgt_name="$1"
	local ig="$2"

	local initiator_name
	local tg_init
	tg_init="$tgt_name "
	for initiator_name in $ig; do
		#
		# escape the periods
		#
		initiator_name=$(echo "$initiator_name" | sed -e 's/\./\\\./g')
		tg_init="${tg_init}${initiator_name},"
	done
	tg_init=$(echo "$tg_init" | sed -e 's/,$//')
	echo "$tg_init"
}

#------------------------------------------------------------------------------
# lock resource creation, timeout the lock if we find a stale lock after 15 sec
# this has to be less than the cluster's  default action timeout value of 50 sec
#------------------------------------------------------------------------------
function lock
{
	#echo "tgt> lock..."
	lockfile -l 15 /tmp/target_lock
}

function unlock
{
	#echo "tgt> ...unlock"
	rm /tmp/target_lock
}

#
# initialize global variables
#
host=$(uname -n)

#
# online or offline, in online mode we require ietd to be running, in
# offline mode we don't interact with ietd
#
run_mode="ietd_offline"

#
# if options are part of a target or the global ietd.conf
# 
option_mode="global"

#
# current target id that either 'target' or 'option' is working on
# this is only usable when the system is in online mode
#
current_tid=0

#
# when generating the conf files, we need to know globally whether the
# user or option commands are running with a current valid target.
# If we find a target that is not destinted for the current node, then
# we must skip the user and option command following the target
#
target_valid=0

cmd="$1"
shift
case "$cmd" in
	configure_cluster)
		configure_targets $*
		;;
	configure_local)
		#
		# configure the node independent target parts
		#
		lock
		option_mode="target"
		configure_localhost_targets $*

		#
		# configure the node dependent target parts (target.allow, for example)
		#
		option_mode="global"
		configure_localhost_iet
		unlock
		;;
	start_internal)
		lock
		option_mode="global"
		configure_localhost_iet
		unlock
		;;
	status)
		$ROOT/iet-tools.plx -c state
		;;
	*)
		echo $"Usage: $0 {status|configure_cluster [online|offline]}"
		exit 1
esac
exit 0
