#!/bin/sh

#
# zfs snapshot/restore support
#
# Steffen Plotner, 2010 - 2011
#

. /etc/iscsi-target/ietd_config.sh

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#
# Functions from here on run only on the current host (the i.e. the host they
# are invoked on, i.e. we have to verify that requested resources are actually
# only when invoked!)
#
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#------------------------------------------------------------------------------
# given the current host, present a list of logical volumes for backup
#------------------------------------------------------------------------------
function zfs_lvlist
{
	local host_current=$(uname -n)
	local vgs=$(vg_active_get $host_current)
	local vg
	for vg in $vgs; do
		cat $VGMETA_DIR/$vg/zfs/backup
	done
}

#------------------------------------------------------------------------------
# validate volume group request (this prevents us having a request come
# into the wrong host)
#------------------------------------------------------------------------------
function zfs_vg_check
{
	local vg_check="$1"

	local host_current=$(uname -n)
	local vgs=$(vg_active_get $host_current)
	local vg
	for vg in $vgs; do
		if [ "$vg" = "$vg_check" ]; then
			echo "$vg"
		fi
	done
}

#------------------------------------------------------------------------------
# on the current host, create a snapshot of a logical volume and prep
# the ddmap data which allows the reader to only read blocks that have
# changed
#------------------------------------------------------------------------------
function zfs_lvsnap
{
	local lun_path="$1"
	local initiator_name="$2"
	local integrity_check="$3"

	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	
	echo "zfs_lvsnap> vg: $vg lv: $lv initiator_name: $initiator_name integrity_check: $integrity_check"
	if [ -z "$vg" -o -z "$lv" -o -z "$initiator_name" ]; then
		echo "zfs_lvsnap_err> zfs_lvsnap: vg/lv initiator_name"
		exit 1
	fi

	local vg_ok=$(zfs_vg_check "$vg")
	if [ -z "$vg_ok" ]; then
		echo "zfs_lvsnap_err> host does not have resource vg: $vg"
		exit 1
	fi

	if [ ! -b "$lun_path" ]; then
		echo "zfs_lvsnap_err> non-existing lun_path: $lun_path"
		exit 1
	fi

	#
	# if there is no snapshot device, then make a snapshot and include ddmap merge
	#
	if [ ! -b ${lun_path}_zfssnap ]; then
	
		#
		# SUSPEND the ddmap and IO operations (at this time both are stopped
		# at the same time, give it 5 seconds so that all worker threads of a target
		# can be put to sleep)
		#
		cmd="$ROOT/ietd_ddmap.sh mapsuspend $lun_path"
		echo "zfs_lvsnap> cmd: $cmd (delay 5 sec)"
		$cmd
		sleep 5

		#
		# create a snaphshot
		#
		ddless_segment_bytes=$(ddless -p | awk -F= '/SEGMENT_SIZE_BYTES/{print $2}')
		ddless_segment_kbytes=$(($ddless_segment_bytes / 1024))
		echo "zfs_lvsnap> snapshot chunk size (ddless segment size): $ddless_segment_kbytes"
		cmd="lvm_snap_create -s $lun_path -c ${ddless_segment_kbytes}k -i 0 -t zfssnap"
		$cmd

		#
		# dump the current ddmap
		#
		cmd="$ROOT/ietd_ddmap.sh ddmap_dump $lun_path"
		echo "zfs_lvsnap> cmd: $cmd"
		$cmd

		#
		# zap the current ddmap
		#
		cmd="$ROOT/ietd_ddmap.sh mapzap $lun_path"
		echo "zfs_lvsnap> cmd: $cmd"
		$cmd

		#
		# RESUME the ddmap and IO operations
		#
		cmd="$ROOT/ietd_ddmap.sh mapresume $lun_path"
		echo "zfs_lvsnap> cmd: $cmd"
		$cmd

		#
		# combine all ddmaps for this lun_path including the state of the ddmap (ie clean or dirty)
		#
		cmd="$ROOT/ietd_ddmap.sh ddmap_combine $lun_path $VGMETA_DIR/$vg/zfs"
		echo "zfs_lvsnap> cmd: $cmd"
		$cmd

		#
		# look for the snapshot blockdevice (might not be immediately online)
		#
		while [ ! -b ${lun_path}_zfssnap ]; do
			echo "zfs_lvsnap> waiting for snap device to come online, sleeping 1 second"
			sleep 1
		done
		
		#
		# make sure we clear out any previous target conf for this snapshot,
		# since we have a new one
		#
		rm $VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf 2> /dev/null
	fi

	#
	# prepare dynamic target
	#
	if [ ! -e "$VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf" ]; then
		cat <<EOF > $VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf
#
# generated by zfs_tools.sh
#
	target up ${lun_path}_zfssnap \\
		blockio DSK-$(date +%y%m%d%H%M%S) $IET_ZFS_VLAN $initiator_name
EOF
	fi

	#
	# activate the target
	#
	$ROOT/ietd_targets.sh configure_local
	
	#
	# determine the name of the created target
	#
	local target=$($ROOT/iet-tools.plx -c lun_path_target -p ${lun_path}_zfssnap)

	#
	# if integrity is asked for, we run the snapshot'd lun through
	# ddless and let it compute a signature, we send the resulting md5
	# via the DATA> channel
	#
	local ddless_md5
	if [ $integrity_check = 1 ]; then
		local tmp_ddless="/tmp/ddless_integrity_$vg_$lv"
		
		cmd="ddless -d -s ${lun_path}_zfssnap -c $tmp_ddless"
		echo "zfs_lvsnap> cmd: $cmd"
		$cmd
		ddless_md5=$(md5sum $tmp_ddless | cut -f1 -d' ')
		
		rm $tmp_ddless
	fi

	echo "DATA> target: $target"
	echo "DATA> ddmap_state: $(cat $VGMETA_DIR/$vg/zfs/${lv}.state)"
	echo "DATA> ddmap_file: $VGMETA_DIR/$vg/zfs/${lv}.ddmap"
	echo "DATA> ddless_md5: $ddless_md5"
}

#------------------------------------------------------------------------------
# remove target and snapshot
#------------------------------------------------------------------------------
function zfs_lvsnapremove
{
	local lun_path="$1"
	local initiator_name="$2"

	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	
	echo "zfs_lvsnapremove> vg: $vg lv: $lv initiator_name: $initiator_name"
	if [ -z "$vg" -o -z "$lv" -o -z "$initiator_name" ]; then
		echo "zfs_lvsnapremove_err> zfs_lvsnap: vg/lv initiator_name"
		exit 1
	fi

	local vg_ok=$(zfs_vg_check "$vg")
	if [ -z "$vg_ok" ]; then
		echo "zfs_lvsnap_err> host does not have resource vg: $vg"
		exit 1
	fi

	if [ ! -b "$lun_path" ]; then
		echo "zfs_lvsnap_err> non-existing lun_path: $lun_path"
		exit 1
	fi

	if [ ! -b "${lun_path}_zfssnap" ]; then
		echo "zfs_lvsnap_err> non-existing lun_path_zfssnap: ${lun_path}_zfssnap"
		exit 1
	fi

	#
	# prepare dynamic target into off mode
	#
	cat "$VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf" | \
		sed -e 's/target up/target down/' > /tmp/$$.conf
	mv /tmp/$$.conf "$VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf"

	#
	# de-activate the target
	#
	$ROOT/ietd_targets.sh configure_local
	rm "$VGMETA_DIR/$vg/targets.d/${lv}_zfssnap.conf"

	#
	# check the snapshot state (i.e. if it got overrun we need to flag the
	# ddmap as dirty)
	#
	local snap_status=$(lvdisplay ${lun_path}_zfssnap | grep "LV snapshot status")
	echo "zfs_lvsnapremove> $snap_status"
	snap_status=$(lvdisplay ${lun_path}_zfssnap | grep "LV snapshot status" | grep "INACTIVE")
	if [ -n "$snap_status" ]; then
		echo "zfs_lvsnapremove> snapshot space was too small, flagging ddmap as dirty"
		ddmap_state_set "$vg/ddmap/$lv" "DDMAP_DIRTY"
	fi

	#
	# remove snapshot
	#
	cmd="lvm_snap_remove -s ${lun_path}_zfssnap -i 0"
	$cmd
}

#------------------------------------------------------------------------------
# make a lun dirty (called by zfs-archive side to inform us that data is not
# consistent after doing an integrity check)
#------------------------------------------------------------------------------
function zfs_lvdirty
{
	local lun_path="$1"
	local initiator_name="$2"

	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	
	echo "zfs_lvdirty> vg: $vg lv: $lv"
	if [ -z "$vg" -o -z "$lv" ]; then
		echo "zfs_lvdirty_err> zfs_lvdirty: vg/lv"
		exit 1
	fi

	local vg_ok=$(zfs_vg_check "$vg")
	if [ -z "$vg_ok" ]; then
		echo "zfs_lvdirty_err> host does not have resource vg: $vg"
		exit 1
	fi

	if [ ! -b "$lun_path" ]; then
		echo "zfs_lvdirty_err> non-existing lun_path: $lun_path"
		exit 1
	fi

	#
	# flag volume as dirty
	#
	ddmap_state_get "$vg/ddmap/$lv"
	ddmap_state_set "$vg/ddmap/$lv" "DDMAP_DIRTY"
	ddmap_state_get "$vg/ddmap/$lv"
}

#------------------------------------------------------------------------------
# create a lun+target for zfs restore purposes
#------------------------------------------------------------------------------
function zfs_restore
{
	local lun_path="$1"
	local lun_size="$2"
	local initiator_name="$3"

	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	
	echo "zfs_restore> vg: $vg lv: $lv lun_size: $lun_size initiator_name: $initiator_name"
	if [ -z "$vg" -o -z "$lv" -o -z "$initiator_name" ]; then
		echo "zfs_restore_err> zfs_restore: vg/lv lun_size initiator_name"
		exit 1
	fi

	local vg_ok=$(zfs_vg_check "$vg")
	if [ -z "$vg_ok" ]; then
		echo "zfs_restore_err> host does not have resource vg: $vg"
		exit 1
	fi

	#
	# do NOT allow LUN overwrite
	#
	if [ -b "$lun_path" ]; then
		echo "zfs_restore_err> existing lun_path: $lun_path, refusing to overwrite"
		exit 1
	fi

	cmd="lvcreate -L$lun_size -n $lv $vg"
	echo "zfs_restore> cmd: $cmd"
	$cmd

	if [ ! -b "$lun_path" ]; then
		echo "zfs_restore_err> lun_path: $lun_path not created?"
		exit 1
	fi

	#
	# prepare dynamic target
	#
	if [ ! -e "$VGMETA_DIR/$vg/targets.d/${lv}_zfsrestore.conf" ]; then
		cat <<EOF > $VGMETA_DIR/$vg/targets.d/${lv}_zfsrestore.conf
#
# generated by zfs_tools.sh
#
	target up ${lun_path} \\
		blockio DSK-$(date +%y%m%d%H%M%S) $IET_ZFS_VLAN $initiator_name
EOF
	fi

	#
	# activate the target
	#
	$ROOT/ietd_targets.sh configure_local
	
	#
	# determine the name of the created target
	#
	local target=$($ROOT/iet-tools.plx -c lun_path_target -p ${lun_path})

	echo "DATA> target: $target"
}

#------------------------------------------------------------------------------
# remove target from zfs restore purposes
#------------------------------------------------------------------------------
function zfs_restoreremove
{
	local lun_path="$1"
	local initiator_name="$2"

	local vg=$(echo "$lun_path" | cut -f3 -d/)
	local lv=$(echo "$lun_path" | cut -f4 -d/)
	
	echo "zfs_restoreremove> vg: $vg lv: $lv initiator_name: $initiator_name"
	if [ -z "$vg" -o -z "$lv" -o -z "$initiator_name" ]; then
		echo "zfs_restoreremove_err> zfs_restoreremove: vg/lv initiator_name"
		exit 1
	fi

	local vg_ok=$(zfs_vg_check "$vg")
	if [ -z "$vg_ok" ]; then
		echo "zfs_restoreremove_err> host does not have resource vg: $vg"
		exit 1
	fi

	if [ ! -b "$lun_path" ]; then
		echo "zfs_restoreremove_err> non-existing lun_path: $lun_path"
		exit 1
	fi

	#
	# prepare dynamic target into off mode
	#
	cat "$VGMETA_DIR/$vg/targets.d/${lv}_zfsrestore.conf" | \
		sed -e 's/target up/target down/' > /tmp/$$.conf
	mv /tmp/$$.conf "$VGMETA_DIR/$vg/targets.d/${lv}_zfsrestore.conf"

	#
	# de-activate the target
	#
	$ROOT/ietd_targets.sh configure_local
	rm "$VGMETA_DIR/$vg/targets.d/${lv}_zfsrestore.conf"
}

cmd="$1"
shift
case "$cmd" in
	vgcheck)
		zfs_vg_check "$1"
		;;
	lvlist)
		zfs_lvlist
		;;
	lvsnap)
		zfs_lvsnap "/dev/$1" "$2" "$3"
		;;
	lvsnapremove)
		zfs_lvsnapremove "/dev/$1" "$2"
		;;
	lvdirty)
		zfs_lvdirty "/dev/$1"
		;;
	lvrestore)
		zfs_restore "/dev/$1" "$2" "$3"
		;;
	lvrestoreremove)
		zfs_restoreremove "/dev/$1" "$2"
		;;
	*)
		echo "Usage: $0 {vgcheck|lvlist|lvsnap|lvsnapremove|lvdirty|lvrestore}"
		exit 1
esac
exit 0
